package net.pingex.dcf.modularity;

import net.pingex.dcf.core.Configuration;
import net.pingex.dcf.core.CoreEventsHandler;
import org.apache.commons.io.FilenameUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.stream.Collectors;

/**
 * Class that load the plugins, a very powerful thingy.
 */
public class PluginLoader
{
	/**
	 * Singleton instance
	 */
	private static PluginLoader INSTANCE = new PluginLoader();

	/**
	 * Logger
	 */
	private static Logger LOGGER = LogManager.getLogger(PluginLoader.class);

	/**
	 * All plugins
	 */
	private Map<String, PluginWrapper> plugins;

	/**
	 * Gives the singleton instance
	 * @return The singleton instance
	 */
	public static PluginLoader getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Enforce singleton with a private constructor.
	 */
	private PluginLoader()
	{
		plugins = new HashMap<>();
	}

	/**
	 * Discover all plugins contained in .jar files in the plugins directory.
	 */
	public void discoverPluginsDirectory()
	{
		LOGGER.debug("Discovering all IPlugin classes in plugins directory.");

		int counter = 0;
		try
		{
			Set<Path> files = Files.walk(Paths.get(Configuration.PLUGINS_DIR))
					.filter(Files::isRegularFile)
					.filter(path -> FilenameUtils.getExtension(path.toString()).equals("jar"))
					.collect(Collectors.toSet());
			for(Path i : files) counter += loadJarFile(i);
		}
		catch(IOException e)
		{
			LOGGER.error("IO error while discovering plugins.", e);
		}

		LOGGER.debug("Loaded {} plugins from {}", counter, Paths.get(Configuration.PLUGINS_DIR).toAbsolutePath());
	}

	/**
	 * Look in a .jar file for IPlugins and preload them.
	 */
	public int loadJarFile(Path target) throws IOException
	{
		LOGGER.debug("Looking up {} for plugins.", target.getFileName());

		// Get DCF-Plugins manifest attribute
		Attributes jarAttributes = new JarFile(target.toFile()).getManifest().getMainAttributes();
		LOGGER.trace("DCF-Plugins attribute: {}", jarAttributes.getValue("DCF-Plugins"));
		if(jarAttributes.getValue("DCF-Plugins") == null)
		{
			LOGGER.debug("Jar file has no `DCF-Plugins` attribute in its manifest. Ignoring archive.");
			return 0;
		}

		String[] pluginsToLoad = jarAttributes.getValue("DCF-Plugins").split(" ");
		LOGGER.debug("Manifest returned {} plugins to load.", pluginsToLoad.length);
		URLClassLoader classLoader = URLClassLoader.newInstance(new URL[]{target.toUri().toURL()});

		int counter = 0;
		for(String item : pluginsToLoad)
		{
			LOGGER.debug("Attempting to preload {}.", item);

			Class<?> clazz;
			try
			{
				clazz = Class.forName(item, true, classLoader);
			}
			catch(ClassNotFoundException e) // Class with that name doesn't exist.
			{
				LOGGER.debug("No class with name {} was found.", item);
				continue;
			}

			// Class doesn't implement IPlugin, or doesn't have @Plugin annotation.
			if(!clazz.isAnnotationPresent(Plugin.class) || !IPlugin.class.isAssignableFrom(clazz))
			{
				LOGGER.debug("Class does not implement IPlugin or doesn't have @Plugin annotation.");
				continue;
			}

			if(plugins.containsKey(clazz.getAnnotation(Plugin.class).id()))
				LOGGER.warn("Found duplicate plugin with id {}. Newly found plugin won't be preloaded.", clazz.getAnnotation(Plugin.class).id());
			else
			{
				try
				{
					PluginWrapper plugin = new PluginWrapper(clazz.asSubclass(IPlugin.class), classLoader);
					plugins.put(clazz.getAnnotation(Plugin.class).id(), plugin);
					LOGGER.debug("Plugin {} version {} preloaded.", clazz.getAnnotation(Plugin.class).id(), clazz.getAnnotation(Plugin.class).version());
					counter++;
					CoreEventsHandler.pluginPreloaded(plugin);
				}
				catch(IllegalAccessException | InstantiationException e)
				{
					LOGGER.warn("Plugin " + clazz.getAnnotation(Plugin.class).id() + " failed to preload.", e);
				}
			}
		}

		LOGGER.debug("Preloaded {} plugins from {}.", counter, target.getFileName());
		return counter;
	}

	/**
	 * Load all plugins that aren't.
	 */
	public void bulkLoadPlugins()
	{
		LOGGER.info("Bulk loading plugins.");
		plugins.entrySet().stream()
				.filter(i -> i.getValue().getState().equals(PluginState.UNLOADED))
				.forEach(i -> loadPlugin(i.getKey()));
	}

	/**
	 * Load a plugin.
	 * @param id The unique ID of the plugin.
	 */
	public void loadPlugin(String id)
	{
		// Check for plugin existence.
		if(!plugins.containsKey(id))
		{
			LOGGER.debug("Plugin with ID {} doesn't exist.", id);
			return;
		}

		LOGGER.debug("Loading plugin {}.", id);
		PluginWrapper plugin = plugins.get(id);

		// Check if plugin state is UNLOADED
		if(!plugin.getState().equals(PluginState.UNLOADED))
		{
			LOGGER.debug("Current plugin state does not allow it to be loaded.");
			return;
		}

		try
		{
			plugin.getInstance().load(Configuration.getStore());
			plugin.setState(PluginState.LOADED);
			LOGGER.info("Loaded plugin {}.", id);
			CoreEventsHandler.pluginLoaded(plugin);
		}
		catch(Throwable throwable)
		{
			LOGGER.warn("Plugin " + id + " failed to load. Entering FAILED state.", throwable);
			plugin.setState(PluginState.FAILED);
		}
	}

	/**
	 * Run all plugins that aren't
	 */
	public void bulkRunPlugins()
	{
		LOGGER.info("Bulk running plugins.");
		plugins.entrySet().stream()
				.filter(i -> i.getValue().getState().equals(PluginState.LOADED) || i.getValue().getState().equals(PluginState.STOPPED))
				.forEach(i -> runPlugin(i.getKey()));
	}

	/**
	 * Run a plugin.
	 * @param id The unique ID of the plugin.
	 */
	public void runPlugin(String id)
	{
		// Check for plugin existence.
		if(!plugins.containsKey(id))
		{
			LOGGER.debug("Plugin with ID {} doesn't exist.", id);
			return;
		}

		LOGGER.debug("Attempting to run plugin {}.", id);
		PluginWrapper plugin = plugins.get(id);

		// Check if plugin state is LOADED or STOPPED
		if(!plugin.getState().equals(PluginState.LOADED) && !plugin.getState().equals(PluginState.STOPPED))
		{
			LOGGER.debug("Current plugin state does not allow it to run.");
			return;
		}

		try
		{
			plugin.getInstance().run();
			LOGGER.info("Plugin {} is now running. (last state: {})", id, plugin.getState());
			plugin.setState(PluginState.RUNNING);
			CoreEventsHandler.pluginRunning(plugin);
		}
		catch(Throwable throwable)
		{
			LOGGER.warn("Plugin " + id + " failed to run. Entering FAILED state. (last state: " + plugin.getState() + ")", throwable);
			plugin.setState(PluginState.FAILED);
		}
	}

	/**
	 * Stop a plugin from running.
	 * @param id The unique ID of the plugin.
	 */
	public void stopPlugin(String id)
	{
		// Check for plugin existence.
		if(!plugins.containsKey(id))
		{
			LOGGER.debug("Plugin with ID {} doesn't exist.", id);
			return;
		}

		LOGGER.debug("Stopping plugin {}.", id);
		PluginWrapper plugin = plugins.get(id);

		// Check if plugin state is RUNNING
		if(!plugin.getState().equals(PluginState.RUNNING))
		{
			LOGGER.debug("Current plugin state does not allow it to stop.");
			return;
		}

		plugin.getInstance().stop();
		LOGGER.info("Stopped plugin {}.", id);
		plugin.setState(PluginState.STOPPED);
		CoreEventsHandler.pluginStopped(plugin);
	}

	/**
	 * Unload a plugin.
	 * @param id The unique ID of the plugin.
	 */
	public void unloadPlugin(String id)
	{
		// Check for plugin existence.
		if(!plugins.containsKey(id))
		{
			LOGGER.debug("Plugin with ID {} doesn't exist.", id);
			return;
		}

		LOGGER.debug("Unloading plugin {}.", id);
		PluginWrapper plugin = plugins.get(id);

		// Check if plugin state is STOPPED
		if(!plugin.getState().equals(PluginState.STOPPED))
		{
			LOGGER.debug("Current plugin state does not allow it to be unloaded.");
			return;
		}

		plugin.getInstance().unload();
		LOGGER.info("Unloaded plugin {}.", id);
		plugin.setState(PluginState.UNLOADED);
		CoreEventsHandler.pluginUnloaded(plugin);
	}

	/**
	 * Get all plugins and their state.
	 */
	public Map<String, PluginWrapper> getPlugins()
	{
		return Collections.unmodifiableMap(plugins);
	}
}